/*
 * esmtp.c
 *
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003	Martin Lambers <marlam@web.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sysexits.h>

#include "merror.h"
#include "net.h"
#include "tls.h"
#include "base64.h"
#include "cram_md5.h"
#include "xmalloc.h"
#include "esmtp.h"


/* according to RFC2822, an SMTP line can contain at most 998 
 * characters + CRLF. We add one character for '\0'.
 * If a line is longer, your MUA (in case of writing) or your
 * providers SMTP sever (in case of reading) violates RFC2822.
 */
#define SMTP_BUFSIZE 1001 	


/*
 * smtp_get_response()
 * 
 * This function gets a message from an smtp server on 'fd'.
 * If 'tls' is an active TLS connection, TLS will be used.
 * Returns the three digit status number. If this number does not equal 
 * 'expected_status', the message from the server is stored via the merror.c
 * functions. Multiline messages are handled. 
 * Return value is the status number or -1 if an error occured.
 */

int smtp_get_response(int fd, tls_t *tls, int expected_status)
{
    int smtp_status;
    char buf[SMTP_BUFSIZE];
    int error_code;
    int error = 0;
    

    if ((error_code = net_getline(fd, tls, buf, SMTP_BUFSIZE)) != 0)
    {
	if (error_code == 1)
	{
	    add_error_line("cannot get a response from SMTP server");
	}
	else /* error_code == 2 */
	{
	    add_error_line("SMTP server sent a line longer than 998 "
		    "characters (this violates RFC 2822)");
	}
	return -1;
    }
    
    smtp_status = atoi(buf);
    
    if (smtp_status != expected_status)
    {
	add_error_line("SMTP error: `%s'", buf);
	error = 1;
    }
    
    /* read continuation lines */
    while (buf[3] == '-')
    {
	if ((error_code = net_getline(fd, tls, buf, SMTP_BUFSIZE)) != 0)
	{
	    if (error_code == 1)
	    {
		add_error_line("cannot get complete response from SMTP "
			"server");
	    }
	    else /* error_code == 2 */
	    {
		add_error_line("SMTP server sent a line longer than 998 "
		     	"characters (this violates RFC 2822)");
	    }
	    return -1;
	}
	if (error)
	{
	    add_error_line("SMTP error (continued): `%s'", buf);
	}
    }

    return smtp_status;
}
	

/*
 * smtp_put()
 *
 * This function is a printf that writes to an SMTP server on 'fd'.
 * If 'tls' is an active TLS connection, TLS will be used. 
 * The string should not end with '\n', because this function appends 
 * an '\r' and an '\n' to it (TCP CRLF).
 * Return value is 1 on write error or buffer overflow and 0 on success.
 */

int smtp_put(int fd, tls_t *tls, char *format, ...)
{
    va_list args;
    int error_code;
    char line[SMTP_BUFSIZE];

    error_code = 0;
    va_start(args, format);
    if (vsnprintf(line, (size_t)(SMTP_BUFSIZE - 3), format, args) 
	    >= SMTP_BUFSIZE - 3)
    {
	/* should never happen */
	add_error_line("line too long in smtp_put()");
	error_code = 1;
    }
    else
    {
	strcat(line, "\r\n");
	if (net_putline(fd, tls, line) != 0)
	{
	    add_error_line("cannot write to SMTP server");
	    error_code = 1;
	}
    }
    va_end(args);
    
    return error_code;
}


/*
 * smtp_starttls()
 *
 * Starts TLS encryption in an SMTP session. 
 * Returns a value from sysexits.h
 */

int smtp_starttls(int fd, tls_t *tls, int tls_method, int nostarttls)
{
    if (!nostarttls)
    {
	if (smtp_put(fd, tls, "STARTTLS") != 0)
	{
	    return EX_IOERR;
	}
	if (smtp_get_response(fd, tls, 220) != 220)
	{
	    add_error_line("line `STARTTLS' not accepted by SMTP server");
	    add_error_line("SMTP server does not accept TLS connections");
	    return EX_UNAVAILABLE;
	}
    }
    if (tls_init(tls, tls_method) != 0)
    {
	add_error_line("cannot initialize TLS");
	return EX_SOFTWARE;
    }
    if (tls_start(tls, fd) != 0)
    {
	add_error_line("cannot start TLS encryption");
	return EX_UNAVAILABLE;
    }

    return 0;
}


/*
 * smtp_auth_plain()
 * 
 * Do SMTP authentication via AUTH PLAIN.
 * Returns a value from sysexits.h
 */

int smtp_auth_plain(int fd, tls_t *tls, char *user, char *password)
{
    char *s;
    char *b64;
    size_t u_len;
    size_t p_len;
    int error_code;

    u_len = strlen(user);
    p_len = strlen(password);
    s = (char *)xmalloc((u_len + p_len + 3) * sizeof(char));
    s[0] = '\0';
    strcpy(s + 1, user);
    strcpy(s + u_len + 2, password);
    b64 = base64enc(s, 1 + u_len + 1 + p_len);
    free(s);
    
    if (smtp_put(fd, tls, "AUTH PLAIN %s", b64) != 0)
    {
	free(b64);
	return EX_IOERR;
    }
    free(b64);
    if ((error_code = smtp_get_response(fd, tls, 235)) != 235)
    {
	if (error_code == 504)
	{
	    add_error_line("AUTH PLAIN not accepted by SMTP server");
	    return EX_UNAVAILABLE;
	}
	else
	{
	    add_error_line("AUTH PLAIN: SMTP server does not "
		"accept username/password");
	    return EX_DATAERR;
	}
    }

    return 0;
}


/*
 * smtp_auth_login()
 * 
 * Do SMTP authentication via AUTH LOGIN.
 * Returns a value from sysexits.h
 */

int smtp_auth_login(int fd, tls_t *tls, char *user, char *password)
{
    char *b64;
    
    if (smtp_put(fd, tls, "AUTH LOGIN") != 0)
    {
	return EX_IOERR;
    }
    if (smtp_get_response(fd, tls, 334) != 334)
    {
	add_error_line("line `AUTH LOGIN' not accepted by SMTP server");
	return EX_UNAVAILABLE;
    }
    b64 = base64enc(user, strlen(user));
    if (smtp_put(fd, tls, "%s", b64) != 0)
    {
	free(b64);
	return EX_IOERR;
    }
    free(b64);
    if (smtp_get_response(fd, tls, 334) != 334)
    {
	add_error_line("AUTH LOGIN: SMTP server does not accept username");
	return EX_DATAERR;
    }
    b64 = base64enc(password, strlen(password));
    if (smtp_put(fd, tls, "%s", b64) != 0)
    {
	free(b64);
	return EX_IOERR;
    }
    free(b64);
    if (smtp_get_response(fd, tls, 235) != 235)
    {
	add_error_line("AUTH LOGIN: SMTP server does not "
		"accept username/password");
	return EX_DATAERR;
    }

    return 0;
}


/*
 * smtp_auth_cram_md5()
 *
 * Do SMTP authentication via AUTH CRAM-MD5.
 * Returns a value from sysexits.h
 */

int smtp_auth_cram_md5(int fd, tls_t *tls, char *user, char *password)
{
    char buffer[SMTP_BUFSIZE];
    int error_code;
    unsigned char digest[16];
    char hex[16] = 
    	{ '0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f' };
    char *p;
    char *b64buf;
    size_t len;
    int i;
    
    if (smtp_put(fd, tls, "AUTH CRAM-MD5") != 0)
    {
	return EX_IOERR;
    }
    if ((error_code = net_getline(fd, tls, buffer, SMTP_BUFSIZE)) != 0)
    {
	if (error_code == 1)
	{
	    add_error_line("cannot get a response from SMTP server");
	}
	else /* error_code == 2 */
	{
	    add_error_line("SMTP server sent a line longer than 998 "
		    "characters (this violates RFC 2822)");
	}
	return EX_IOERR;
    }
    if (strncmp(buffer, "334 ", 4) != 0)
    {
	add_error_line("SMTP error: `%s'", buffer);
	add_error_line("line `AUTH CRAM-MD5' not accepted by SMTP server");
	return EX_UNAVAILABLE;
    }
    b64buf = base64dec(buffer + 4, &len);
    if (!b64buf)
    {
	/* should never happen */
	add_error_line("AUTH CRAM-MD5: server sent invalid challenge");
	return EX_PROTOCOL;
    }
    cram_md5(password, strlen(password), b64buf, len, digest);
    free(b64buf);
    if (strlen(user) + 1 + 33 > SMTP_BUFSIZE)
    {
	/* should never happen */
	add_error_line("username too long for AUTH CRAM-MD5 authentication");
	return EX_DATAERR;
    }
    /* buffer is long enough */
    strcpy(buffer, user);
    p = buffer + strlen(user);
    *p++ = ' ';
    for (i = 0; i < 16; i++)
    {
	p[2 * i] = hex[digest[i] >> 4];
	p[2 * i + 1] = hex[digest[i] & 0x0f];
    }
    p[32] = '\0';
    p = base64enc(buffer, strlen(buffer));
    if (strlen(p) >= SMTP_BUFSIZE)
    {
	/* should never happen */
	free(p);
	add_error_line("username too long for AUTH CRAM-MD5 authentication");
	return EX_DATAERR;
    }
    /* buffer is long enough */
    strcpy(buffer, p);
    free(p);
    if (smtp_put(fd, tls, "%s", buffer) != 0)
    {
	return EX_IOERR;
    }
    if (smtp_get_response(fd, tls, 235) != 235)
    {
	add_error_line("AUTH CRAM-MD5: SMTP server does not "
		"accept username/password");
	return EX_DATAERR;
    }

    return 0;
}


/*
 * smtp_send()
 *
 * Send the mail (read from 'mailf') to the SMTP server on 'server:port'.
 * Authenticate with AUTH LOGIN if 'user' is not NULL ('password' must not be
 * NULL in this case of course).
 * 'tls_method' is the TLS method to use, see tls.h
 * The number of recipients is stored in 'rcptc', the recipient mail addresses 
 * in 'rcptv'. The 'from' string should contain the mail address of the user to
 * be accepted by the server. 
 * The Bcc header will be removed from the mail and not sent to the SMTP
 * server unless the 'keep_bcc' flag is set.
 * The server is assumed to use SSL/TLS exclusively (no STARTTLS command
 * needed) if the 'nostarttls' flag is set.
 * The function returns 0 if the message was successfully send, and a value
 * from sysexits.h indicating the error on failure.
 */

int smtp_send(char *server, int port, char *from, 
	int auth_method, char *user, char *password,
	int tls_method, int nostarttls, int keep_bcc,
	int rcptc, char *rcptv[], 
	FILE *mailf)
{
    char buffer[SMTP_BUFSIZE];
    char *p;
    int error_code;
    int i;
    int fd;
    tls_t tls;
    int in_header;
    int in_bcc;
    
    tls_clear(&tls);
    /* connect to smtp server */
    if ((fd = open_socket(server, port)) < 0)
    {
	add_error_line("cannot connect to SMTP server");
	error_code = EX_UNAVAILABLE;
	goto exit;
    }
    /* start TLS encryption for TLS-only servers (no STARTTLS) if wanted */
    if (tls_method != TLS_METHOD_NONE && nostarttls)
    {
	if ((error_code = smtp_starttls(fd, &tls, tls_method, nostarttls)) != 0)
	{
	    add_error_line("TLS encryption failed");
	    goto close_exit;
	}
    }
    /* get initial ok message */
    if (smtp_get_response(fd, &tls, 220) != 220)
    {
	add_error_line("cannot get initial ok message from SMTP server");
	error_code = EX_UNAVAILABLE;
	goto close_exit;
    }
    /* start smtp session */
    if (smtp_put(fd, &tls, "EHLO localhost") != 0)
    {
	error_code = EX_IOERR;
	goto close_exit;
    }
    if (smtp_get_response(fd, &tls, 250) != 250)
    {
	add_error_line("line `EHLO localhost' not accepted by SMTP server");
	error_code = EX_DATAERR;
	goto close_exit;
    }
    /* start TLS encryption with STARTTLS if wanted */
    if (tls_method != TLS_METHOD_NONE && !nostarttls)
    {
	if ((error_code = smtp_starttls(fd, &tls, tls_method, nostarttls)) != 0)
	{
	    add_error_line("TLS encryption failed");
	    goto close_exit;
	}
    }
    /* do SMTP AUTH authentication if wanted */
    if (auth_method != AUTH_METHOD_NONE)
    {
	if (auth_method == AUTH_METHOD_CRAM_MD5)
	{
	    if ((error_code = smtp_auth_cram_md5(fd, &tls, user, password)) != 0)
	    {
		add_error_line("AUTH CRAM-MD5 authentication failed");
		goto close_exit;
	    }
	}
	else if (auth_method == AUTH_METHOD_PLAIN)
	{
	    if ((error_code = smtp_auth_plain(fd, &tls, user, password)) != 0)
	    {
		add_error_line("AUTH PLAIN authentication failed");
		goto close_exit;
	    }
	}
	else if (auth_method == AUTH_METHOD_LOGIN)
	{
	    if ((error_code = smtp_auth_login(fd, &tls, user, password)) != 0)
	    {
		add_error_line("AUTH LOGIN authentication failed");
		goto close_exit;
	    }
	}
	else
	{
	    add_error_line("BUG: unknown SMTP AUTH method in smtp_send()");
	    error_code = EX_SOFTWARE;
	    goto close_exit;
	}
    }
    /* tell the server who wants to send mail and to whom */
    if (smtp_put(fd, &tls, "MAIL FROM: %s", from) != 0)
    {
	error_code = EX_IOERR;
	goto close_exit;
    }
    if (smtp_get_response(fd, &tls, 250) != 250)
    {
	add_error_line("line `MAIL FROM: %s' not accepted by SMTP server", 
		from);
	error_code = EX_DATAERR;
	goto close_exit;
    }
    for (i = 0; i < rcptc; i++)
    {
	if (smtp_put(fd, &tls, "RCPT TO: %s", rcptv[i]) != 0)
	{
	    error_code = EX_IOERR;
	    goto close_exit;
	}
	if (smtp_get_response(fd, &tls, 250) != 250)
	{
	    add_error_line("line `RCPT TO: %s' not accepted by SMTP server", 
		    rcptv[i]);
	    error_code = EX_DATAERR;
	    goto close_exit;
	}
    }
    /* send the mail */
    if (smtp_put(fd, &tls, "DATA") != 0)
    {
	error_code = EX_IOERR;
	goto close_exit;
    }
    if (smtp_get_response(fd, &tls, 354) != 354)
    {
	add_error_line("SMTP server does not accept the mail");
	error_code = EX_DATAERR;
	goto close_exit;
    }
    /* Copy the mail from mailf to fd, converting '\n' to '\r\n' and a leading
     * dot on a line to two leading dots; the smtp server deletes a leading
     * dot (RFC 2821, section 4.5.2).
     * Strip the Bcc header unless the keep_bcc flag is set. */
    in_header = 1;
    in_bcc = 0;
    while (fgets(buffer, (int)sizeof(buffer), mailf) != NULL)
    {
	p = strchr(buffer, '\n');
	if (!p)
	{
	    add_error_line("line longer than 998 characters; "
		    "your MUA violates RFC 2822!");
	    error_code = EX_DATAERR;
	    goto close_exit;
	}
	*p = '\0';
	if (!keep_bcc)
	{
	    if (in_header && buffer[0] == '\0')
	    {
		in_header = 0;
	    }
	    if (in_header)
	    {
		if (!in_bcc)
		{
		    if (strlen(buffer) >= 5
			    && strncmp(buffer, "Bcc: ", (size_t)5) == 0)
		    {
			in_bcc = 1;
			/* remove Bcc header by ignoring this line */
			continue;
		    }
		}
		else
		{
		    /* continued header lines begin with "horizontal whitespace"
		     * (RFC 2822, section 2.2.3) */
		    if (buffer[0] == '\t' || buffer[0] == ' ')
		    {
			/* remove continued Bcc header by ignoring this line */
			continue;
		    }
		    else
		    {
			in_bcc = 0;
		    }
		}
	    }
	}
	if (buffer[0] == '.')
	{
	    /* This works because we turned '\n\0' to '\0\0'. */
	    memmove(buffer + 1, buffer, strlen(buffer) * sizeof(char));
	    buffer[0] = '.';
	}
	/* smtp_put will append "\r\n" */
	if (smtp_put(fd, &tls, "%s", buffer) != 0)
	{
	    error_code = EX_IOERR;
	    goto close_exit;
	}
    }
    if (ferror(mailf))
    {
	add_error_line("input error while reading the mail");
	error_code = EX_IOERR;
	goto close_exit;
    }
    /* end the mail */
    if (smtp_put(fd, &tls, ".") != 0)
    {
	error_code = EX_IOERR;
	goto close_exit;
    }
    if (smtp_get_response(fd, &tls, 250) != 250)
    {
	add_error_line("SMTP server did not accept the mail");
	error_code = EX_DATAERR;
	goto close_exit;
    }
    
    /* Since the smtp server responded with 250, we can be sure that the mail
     * will be delivered. Our job is done. We can ignore following i/o and
     * smtp errors. */
    error_code = 0;
    smtp_put(fd, &tls, "QUIT");
    smtp_get_response(fd, &tls, 221);

close_exit:
    /* close connection and clean up */
    tls_close(&tls);
    close(fd);

exit:
    return error_code;
}
