/*
 * tls.c
 *
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003	Martin Lambers <marlam@web.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <limits.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>

#include "merror.h"
#include "tls.h"


/*
 * Always use tls_clear() or tls_init()/tls_start() before using a tls_t!
 * Never call a tls_*() function with tls_t NULL!
 */


/*
 * tls_clear()
 *
 * Clears a tls_t type (marks it inactive).
 */

void tls_clear(tls_t *tls)
{
    tls->ssl_ctx = NULL;
    tls->ssl = NULL;
}


/*
 * tls_init()
 *
 * Initialize SSL library and an SSL_CTX for a client connection.
 * 'method' must be one of the methods defined in tls.h
 * Returns 0 on success and 1 on error.
 */

int tls_init(tls_t *tls, int method)
{
    SSL_METHOD *ssl_method = NULL;
    
    SSL_load_error_strings();
    SSL_library_init();
    if (!RAND_status)
    {
	/* FIXME: seed random number generator 
	 * On systems whith /dev/urandom, OpenSSL seeds itself.
	 * On other systems, we have to do it ourselves with RAND_seed().
	 * I have no idea with what. */
	add_error_line("BUG: must seed random number generator for TLS [%s]", 
		ERR_error_string(ERR_get_error(), NULL));
	return 1;
    }
    switch (method)
    {
	case TLS_METHOD_CHOOSE:
	    ssl_method = SSLv23_client_method();
	    break;

	case TLS_METHOD_SSL2:
	    ssl_method = SSLv2_client_method();
	    break;
	    
	case TLS_METHOD_SSL3:
	    ssl_method = SSLv3_client_method();
	    break;

	case TLS_METHOD_TLS1:
	    ssl_method = TLSv1_client_method();
	    break;

	default:
	    add_error_line("BUG: unknown TLS method in init_tls()");
	    return 1;
    }
    if (!ssl_method)
    {
	add_error_line("TLS method %d failed", method);
	return 1;
    }
    tls->ssl_ctx = SSL_CTX_new(ssl_method);
    if (!tls->ssl_ctx)
    {
	add_error_line("cannot create TLS context [%s]", 
		ERR_error_string(ERR_get_error(), NULL));
	return 1;
    }

    return 0;
}


/*
 * tls_is_active()
 *
 * Returns 1 if 'tls' is an active TLS connection, otherwise 0.
 */

int tls_is_active(tls_t *tls)
{
    return (tls->ssl != NULL);
}


/*
 * tls_io_error()
 *
 * Add an error line according to 'error_code' (which originates from an
 * SSL_read(), SSL_write() or SSL_connect() operation) and 'error_code2' (which
 * originates from an SSL_get_error() call with 'error_code' as its argument).
 * The line will read: "error_string [error_reason]". 'error_string' is given by
 * the calling function, this function finds out 'error_reason'.
 */

void tls_io_error(int error_code, int error_code2, char *error_string)
{
    unsigned long error_code3;
    char *error_reason;
    
    switch (error_code2)
    {
	case SSL_ERROR_NONE:
	    error_reason = "unknown error";
	    break;
		
	case SSL_ERROR_SYSCALL:
	    error_code3 = ERR_get_error();
    	    if (error_code3 == 0)
	    {
		if (error_code == 0)
		{
		    error_reason = "a protocol violating EOF occured";
		}
		else if (error_code == -1)
		{
		    error_reason = strerror(errno);
		}
		else
		{
		    error_reason = "unknown error";
		}
	    }
	    else
	    {
		error_reason = ERR_error_string(error_code3, NULL);
	    }
	    break;
	    
	case SSL_ERROR_ZERO_RETURN:
	    error_reason = "the connection has been closed unexpectedly";
	    break;

	case SSL_ERROR_SSL:
	    error_reason = ERR_error_string(ERR_get_error(), NULL);
	    break;
	    
	default:
    	    error_reason = "unknown error";
	    break;
    }
    add_error_line("%s [%s]", error_string, error_reason);
}


/*
 * tls_start
 *
 * Starts TLS encryption on a socket.
 * Returns 0 on success and 1 on error.
 */

int tls_start(tls_t *tls, int fd)
{
    int error_code;
    
    if (!(tls->ssl = SSL_new(tls->ssl_ctx)))
    {
	add_error_line("SSL_new() failed [%s]", 
		ERR_error_string(ERR_get_error(), NULL));
	return 1;
    }
    if (!SSL_set_fd(tls->ssl, fd))
    {
	add_error_line("SSL_set_fd() failed [%s]", 
		ERR_error_string(ERR_get_error(), NULL));
	SSL_free(tls->ssl);
	tls->ssl = NULL;
	return 1;
    }
    if ((error_code = SSL_connect(tls->ssl)) < 1)
    {
	tls_io_error(error_code, SSL_get_error(tls->ssl, error_code),
		"cannot establish TLS connection");
	SSL_free(tls->ssl);
	tls->ssl = NULL;
	return 1;
    }
    if (!SSL_get_peer_certificate(tls->ssl))
    {
	add_error_line("cannot get server certificate [%s]", 
		ERR_error_string(ERR_get_error(), NULL));
	SSL_free(tls->ssl);
	tls->ssl = NULL;
	return 1;
    }
    /* TODO: Do something with the server certificate */

    return 0;
}


/*
 * tls_getchar()
 *
 * Reads a character using TLS and stores it in 'c'.
 * Returns 0 on success, -1 on EOF and 1 on error.
 */

int tls_getchar(tls_t *tls, char *c)
{
    int error_code;
    int error_code2;
    
    if ((error_code = SSL_read(tls->ssl, c, 1)) < 1)
    {
	if ((error_code2 = SSL_get_error(tls->ssl, error_code)) 
		== SSL_ERROR_NONE)
	{
	    /* EOF */
	    return -1;
	}
	else
	{
	    tls_io_error(error_code, error_code2, 
		    "cannot read from TLS connection");
	    return 1;
	}
    }
    else
    {
	return 0;
    }
}


/*
 * tls_putline()
 *
 * Writes a line using TLS.
 * Returns 0 on success and 1 on error.
 */

int tls_putline(tls_t *tls, char *line)
{
    int error_code;
    size_t count;

    count = strlen(line);
    if (count > (size_t)INT_MAX)
    {
	add_error_line("line too long in tls_putline()");
	return 1;
    }
    else if (count < 1)
    {
	/* nothing to be done */
	return 0;
    }
    
    if ((error_code = SSL_write(tls->ssl, line, (int)count)) != (int)count)
    {
	tls_io_error(error_code, SSL_get_error(tls->ssl, error_code),
		"cannot write to TLS connection");
	return 1;
    }

    return 0;
}


/*
 * tls_close()
 *
 * Close 'tls->ssl', free 'tls->ssl' and 'tls->ssl_ctx'
 */

void tls_close(tls_t *tls)
{
    if (tls->ssl)
    {
	SSL_shutdown(tls->ssl);
	SSL_free(tls->ssl);
    }
    if (tls->ssl_ctx)
    {
	SSL_CTX_free(tls->ssl_ctx);
    }
}
